from typing import TypeVar

from .tables import Table

GenericTable = TypeVar("GenericTable", bound=Table)


def defragment(table: GenericTable) -> GenericTable:
    """Condense the underlying memory which backs the table to make
    it all contiguous. This makes many operations more efficient after
    defragmentation is complete.

    """

    combined = table.table.combine_chunks()
    return table.__class__(table=combined)



from .__version__ import __version__
from .attributes import FloatAttribute, IntAttribute, StringAttribute
from .columns import (
    BinaryColumn,
    Column,
    Date32Column,
    Date64Column,
    Decimal128Column,
    Decimal256Column,
    DictionaryColumn,
    DurationColumn,
    Float16Column,
    Float32Column,
    Float64Column,
    Int8Column,
    Int16Column,
    Int32Column,
    Int64Column,
    LargeBinaryColumn,
    LargeListColumn,
    LargeStringColumn,
    ListColumn,
    MapColumn,
    MonthDayNanoIntervalColumn,
    NullColumn,
    RunEndEncodedColumn,
    StringColumn,
    StructColumn,
    SubTableColumn,
    Time32Column,
    Time64Column,
    TimestampColumn,
    UInt8Column,
    UInt16Column,
    UInt32Column,
    UInt64Column,
)
from .concat import concatenate
from .defragment import defragment
from .errors import InvariantViolatedError, TableFragmentedError, ValidationError
from .indexing import StringIndex
from .matrix import MatrixArray, MatrixExtensionType
from .tables import Table
from .validators import and_, eq, ge, gt, is_in, le, lt

__all__ = [
    "__version__",
    "Table",
    "MatrixArray",
    "MatrixExtensionType",
    "concatenate",
    "StringIndex",
    "Column",
    "SubTableColumn",
    "Int8Column",
    "Int16Column",
    "Int32Column",
    "Int64Column",
    "UInt8Column",
    "UInt16Column",
    "UInt32Column",
    "UInt64Column",
    "Float16Column",
    "Float32Column",
    "Float64Column",
    "StringColumn",
    "LargeBinaryColumn",
    "LargeStringColumn",
    "Date32Column",
    "Date64Column",
    "TimestampColumn",
    "Time32Column",
    "Time64Column",
    "DurationColumn",
    "MonthDayNanoIntervalColumn",
    "BinaryColumn",
    "Decimal128Column",
    "Decimal256Column",
    "NullColumn",
    "ListColumn",
    "LargeListColumn",
    "MapColumn",
    "DictionaryColumn",
    "StructColumn",
    "RunEndEncodedColumn",
    "ValidationError",
    "TableFragmentedError",
    "InvariantViolatedError",
    "lt",
    "le",
    "gt",
    "ge",
    "eq",
    "and_",
    "is_in",
    "StringAttribute",
    "IntAttribute",
    "FloatAttribute",
    "defragment",
]



from typing import Callable, Optional

import pyarrow as pa


def _walk_schema(
    field: pa.Field,
    visitor: Callable[[pa.Field, list[pa.Field]], None],
    ancestors: Optional[list[pa.Field]] = None,
) -> None:
    # Visit every struct field embedded within a field in depth-first fashion.
    #
    # Runs visitor(struct_field, [ancestors...]) on each struct field.
    if ancestors is None:
        ancestors = [field]
    else:
        ancestors += [field]

    for subfield in field.type:
        if pa.types.is_struct(subfield.type):
            _walk_schema(subfield, visitor, ancestors)

    ancestors.pop()
    visitor(field, ancestors)



from typing import Iterator, TypeVar

import pyarrow as pa

from .defragment import defragment
from .tables import Table

T = TypeVar("T", bound=Table)


def concatenate(values: Iterator[T], defrag: bool = True) -> T:
    """Concatenate a collection of Tables into a single Table.

    All input Tables must have the same schema (typically, this will
    be from being the same class).

    By default, results are compacted to be contiguous in memory,
    which involves a copy. In a tight loop, this can be very
    inefficient, so you can set the 'defrag' parameter to False to
    skip this compaction step, and instead call quiver.defragment on
    the result after the loop is complete.

    """
    batches = []
    first = True
    for v in values:
        batches += v.table.to_batches()
        if first:
            cls = v.__class__
            first = False
    table = pa.Table.from_batches(batches)
    result = cls(table=table)
    if defrag:
        result = defragment(result)
    return result



from typing import Any

import numpy.typing as npt
import pyarrow as pa


class MatrixExtensionType(pa.PyExtensionType):  # type: ignore
    """This is a custom type for embedding multi-dimensional arrays
    into Table schemas.

    """

    def __init__(self, shape: tuple[int, int], element_dtype: pa.DataType):
        self.shape = shape
        self.element_dtype = element_dtype
        self.dtype = element_dtype
        for dim in reversed(self.shape):
            self.dtype = pa.list_(self.dtype)
        pa.PyExtensionType.__init__(self, self.dtype)

    def __reduce__(self) -> tuple[type, tuple[tuple[int, int], pa.DataType]]:
        return (
            MatrixExtensionType,
            (self.shape, self.element_dtype),
        )

    def __arrow_ext_class__(self) -> type:
        return MatrixArray


class MatrixArray(pa.ExtensionArray):  # type: ignore
    def to_numpy(self) -> npt.NDArray[Any]:
        storage = self.storage
        row_size = 1
        for dim in reversed(self.type.shape):
            storage = storage.flatten()
            row_size *= dim
        np_array = storage.to_numpy()

        n_row = len(storage) // row_size
        reshaped: npt.NDArray[Any] = np_array.reshape((n_row, *self.type.shape))
        return reshaped



import struct
from typing import TYPE_CHECKING, Generic, Optional, TypeVar

if TYPE_CHECKING:
    from .tables import Table

T = TypeVar("T")


class Attribute(Generic[T]):
    """
    An Attribute is an accessor for scalar data in a table.

    It works through the descriptor protocol.

    All attributes must be serializable into bytes.

    Attributes are stored as metadata on the underlying Arrow table's schema.
    """

    _type: type[T]

    def __init__(
        self,
        default: Optional[T] = None,
    ):
        self.default = default
        self.name = "__ERR_UNSET_NAME"

    def __get__(self, instance: "Table", owner: type) -> T:
        if instance is None:
            return self
        name = self.name.encode("utf8")
        if instance.table.schema.metadata is None or name not in instance.table.schema.metadata:
            if self.default is not None:
                return self.default
            raise AttributeError(f"Attribute {self.name} is not set and has no default")
        raw = instance.table.schema.metadata[name]
        return self.from_bytes(raw)

    def __set__(self, instance: "Table", value: T) -> None:
        metadata = instance.table.schema.metadata
        if metadata is None:
            metadata = {}
        metadata[self.name.encode("utf8")] = self.to_bytes(value)
        instance.table = instance.table.replace_schema_metadata(metadata)

    def __set_name__(self, owner: type, name: str) -> None:
        self.name = name

    def to_bytes(self, value: T) -> bytes:
        """
        Convert the value to bytes.
        """
        raise NotImplementedError

    def from_bytes(self, raw: bytes) -> T:
        """
        Convert the bytes to a value.
        """
        raise NotImplementedError

    def to_string(self, value: T) -> str:
        """
        Convert the value to a string. Used for CSV writing.
        """
        raise NotImplementedError

    def from_string(self, raw: str) -> T:
        """
        Convert a string to a value. Used for CSV reading.
        """
        raise NotImplementedError


class StringAttribute(Attribute[str]):
    _type = str

    def __init__(self, default: Optional[str] = None):
        super().__init__(default=default)

    def to_bytes(self, value: str) -> bytes:
        return value.encode("utf8")

    def from_bytes(self, raw: bytes) -> str:
        return raw.decode("utf8")

    def to_string(self, value: str) -> str:
        return value

    def from_string(self, raw: str) -> str:
        return raw


class IntAttribute(Attribute[int]):
    _type = int

    def __init__(self, default: Optional[int] = None, nbytes: int = 8, signed: bool = True):
        self.nbytes = nbytes
        self.signed = signed
        super().__init__(default=default)

    def to_bytes(self, value: int) -> bytes:
        return value.to_bytes(length=self.nbytes, byteorder="little", signed=self.signed)

    def from_bytes(self, raw: bytes) -> int:
        return int.from_bytes(raw, byteorder="little", signed=self.signed)

    def to_string(self, value: int) -> str:
        return str(value)

    def from_string(self, raw: str) -> int:
        return int(raw)


class FloatAttribute(Attribute[float]):
    _type = float

    def __init__(self, default: Optional[float] = None, nbytes: int = 8):
        if nbytes == 8:
            self._struct_fmt = "<d"
        elif nbytes == 4:
            self._struct_fmt = "<f"
        elif nbytes == 2:
            self._struct_fmt = "<e"
        else:
            raise ValueError("nbytes must be 2, 4 or 8")
        super().__init__(default=default)

    def to_bytes(self, value: float) -> bytes:
        return struct.pack(self._struct_fmt, value)

    def from_bytes(self, raw: bytes) -> float:
        return float(struct.unpack(self._struct_fmt, raw)[0])

    def to_string(self, value: float) -> str:
        return "{:.17g}".format(value)

    def from_string(self, raw: str) -> float:
        return float(raw)
